/**
 * Copyright 2019 AXA Group Operations S.A.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import {
  BoundingBox,
  Document,
  Element,
  Font,
  Heading,
  Line,
  Page,
  Paragraph,
  Word,
} from '../../types/DocumentRepresentation';
import * as utils from '../../utils';
import logger from '../../utils/Logger';
import { LinesToParagraphModule } from '../LinesToParagraphModule/LinesToParagraphModule';
import { Module } from '../Module';

export class HeadingDetectionModule extends Module {
  public static moduleName = 'heading-detection';
  public static dependencies = [LinesToParagraphModule];

  public main(doc: Document): Document {
    if (this.headingsDetected(doc)) {
      logger.warn(
        'Warning: this page already has some headings in it. Not performing heading detection.',
      );
      return doc;
    }

    // get the main body font from all the words in document
    const mainCommonFont = this.commonFont(doc);

    // get the main body font from all the words in document
    // const mainCommonLineHeight = this.commonLineHeight(doc);

    // get all the fonts to use for heading
    const headingFonts = this.heagingFonts(doc);

    doc.pages.forEach((page: Page) => {
      const newContents = this.extractHeadings(
        page,
        mainCommonFont,
        headingFonts,
        // mainCommonLineHeight,
      );
      const paras: Paragraph[] = this.mergeLinesIntoParagraphs(newContents.paragraphLines);
      const headings: Heading[] = this.mergeLinesIntoHeadings(newContents.headingLines);
      page.elements = newContents.rootElements.concat([...headings, ...paras]);
    });

    if (this.headingsDetected(doc)) {
      this.computeHeadingLevels(doc);
    }
    return doc;
  }

  /**
   * Takes into account potential headings inside a paragraph
   * splits a paragraph into multiple ones and returns heading candidates
   * @param page Page to extract headings from each paragraph
   * @param commonFont Most used font in document
   * @param headingFonts Array of only fonts that all headings should use
   */
  private extractHeadings(
    page: Page,
    commonFont: Font,
    headingFonts: Font[],
    // commonLineHeight: number,
  ): { headingLines: Line[][]; paragraphLines: Line[][]; rootElements: Element[] } {
    // get all paragraphs in page root
    const pageParagraphs = this.pageParagraphs(page);
    return {
      ...this.extractHeadingsInParagraphs(
        pageParagraphs,
        commonFont,
        headingFonts,
        // commonLineHeight,
      ),
      rootElements: this.createHeadingsInOtherElements(
        this.pageOtherElements(page),
        commonFont,
        headingFonts,
        // commonLineHeight,
      ),
    };
  }

  private createHeadingsInOtherElements(
    elements: Element[],
    commonFont: Font,
    headingFonts: Font[],
    // commonLineHeight: number,
  ): Element[] {
    this.getElementsWithParagraphs(elements, []).forEach(element => {
      const newContents = this.extractHeadingsInParagraphs(
        this.getParagraphsInElement(element),
        commonFont,
        headingFonts,
        // commonLineHeight,
      );
      const paras: Paragraph[] = this.mergeLinesIntoParagraphs(newContents.paragraphLines);
      const headings: Heading[] = this.mergeLinesIntoHeadings(newContents.headingLines);

      element.content = [...headings, ...paras];
    });
    return elements;
  }

  private extractHeadingsInParagraphs(
    paragraphs: Paragraph[],
    commonFont: Font,
    headingFonts: Font[],
    // commonLineHeight: number,
  ): { headingLines: Line[][]; paragraphLines: Line[][] } {
    const paragraphLines: Line[][] = [];
    const headingLines: Line[][] = [];

    paragraphs
      .map(paragraph => paragraph.content)
      .forEach(linesInParagraph => {
        let initiatedHeading = false;
        if (commonFont instanceof Font) {
          const headingIdx: number[] = linesInParagraph
            .map((line: Line, pos: number) => {
              if (
                (pos === 0 || initiatedHeading) &&
                this.isHeadingCandidate(
                  line,
                  commonFont,
                  headingFonts,
                  // commonLineHeight,
                  // utils.isGeneralUpperCase(linesInParagraph),
                  // utils.isGeneralTitleCase(linesInParagraph),
                )
              ) {
                initiatedHeading = true;
                return pos;
              } else {
                initiatedHeading = false;
                return undefined;
              }
            })
            .filter((i: number) => i !== undefined);
          if (headingIdx.length > 0) {
            const lineIdx: number[] = [...Array(linesInParagraph.length).keys()].filter(
              x => !headingIdx.includes(x),
            );
            utils.groupConsecutiveNumbersInArray(lineIdx).forEach((group: number[]) => {
              const newLines: Line[] = [];
              group.forEach((id: number) => {
                newLines.push(linesInParagraph[id]);
              });
              if (newLines.length > 0) {
                paragraphLines.push(newLines);
              }
            });
            this.groupHeadingsByFont(headingIdx, linesInParagraph).forEach(headingGroup => {
              utils.groupConsecutiveNumbersInArray(headingGroup).forEach((group: number[]) => {
                const newHeadings: Line[] = [];
                group.forEach((id: number) => {
                  newHeadings.push(linesInParagraph[id]);
                });
                if (newHeadings.length > 0) {
                  headingLines.push(newHeadings);
                }
              });
            });
          } else {
            paragraphLines.push(linesInParagraph);
          }
        } else {
          logger.warn("can't account for headings while para merge - no font info available");
        }
      });

    if (headingLines.length > 0) {
      return { headingLines, paragraphLines };
    }

    return {
      headingLines: [],
      paragraphLines: paragraphs.map(paragraph => paragraph.content),
    };
  }

  private isHeadingCandidate(
    line: Line,
    mostCommonFont: Font,
    headingFonts: Font[],
    // commonLineHeight: number,
    // generalUpperCase: boolean,
    // generalTitleCase: boolean,
  ): boolean {
    const serializedHeadingFonts = headingFonts.map(font => JSON.stringify(font));
    if (!serializedHeadingFonts.includes(JSON.stringify(this.noColourFont(line.getMainFont())))) {
      return false;
    }
    return this.isHeadingLine(line, mostCommonFont); // , commonLineHeight); // , generalUpperCase, generalTitleCase);
  }

  private isHeadingLine(
    line: Line,
    commonFont: Font,
    // commonLineHeight: number,
    // upperCase: boolean,
    // titleCase: boolean,
  ): boolean {
    // console.log(upperCase + ' ' + titleCase + ' ' + line.parent);
    // const isBigger = Math.round(line.getMainFont().size) > Math.round(commonFont.size);

    // const isLineHeightBigger = line.height > commonLineHeight;
    const isDifferentStyle = line.getMainFont().weight !== commonFont.weight;
    // return isBigger || isLineHeightBigger || (isDifferentStyle && line.isUniqueFont());
    const isFontBigger = line.getMainFont().size > commonFont.size;
    // const lineHeightFontRatio = line.height / line.getMainFont().size;

    return isFontBigger || (isDifferentStyle && line.isUniqueFont()); // || lineHeightFontRatio > 1.0;
    // return isFontBigger || (isDifferentStyle && line.isUniqueFont());
    /*return (
      line.getMainFont().size > commonFont.size ||
      (line.getMainFont().weight === 'bold' && commonFont.weight !== 'bold') ||
      (line.content.map(w => RegExp(/^[a-z][A-z]*$/gm).test(w.toString())).filter(p => p).length >
        0 &&
        (line.toString().toUpperCase() === line.toString() && !upperCase)) ||
      (line.content.map(w => RegExp(/^[a-z][A-z]*$/gm).test(w.toString())).filter(p => p).length >
        0 &&
        (utils.toTitleCase(line.toString()) === line.toString() && !titleCase))
    );*/
  }

  private groupHeadingsByFont(headingIndexes: number[], lines: Line[]): number[][] {
    // Skip join heading lines if they doesn't have same font
    const fontGroupedHeadings: number[][] = [];
    let joinedHeadings: number[] = [];
    headingIndexes.forEach((pos, index) => {
      const currentHeadingLineFont = lines[pos].getMainFont();
      const prevHeadingLineFont = index > 0 ? lines[index - 1].getMainFont() : null;
      if (!prevHeadingLineFont || currentHeadingLineFont.isEqual(prevHeadingLineFont)) {
        joinedHeadings.push(pos);
      } else {
        fontGroupedHeadings.push(joinedHeadings);
        joinedHeadings = [pos];
      }
    });
    if (joinedHeadings.length > 0) {
      fontGroupedHeadings.push(joinedHeadings);
    }

    return fontGroupedHeadings;
  }

  private mergeLinesIntoParagraphs(joinedLines: Line[][]): Paragraph[] {
    return joinedLines.map((group: Line[]) => {
      const paragraph: Paragraph = utils.mergeElements<Line, Paragraph>(
        new Paragraph(BoundingBox.merge(group.map((l: Line) => l.box))),
        ...group,
      );
      paragraph.properties.order = group[0].properties.order;
      return paragraph;
    });
  }

  private mergeLinesIntoHeadings(joinedLines: Line[][]): Heading[] {
    return joinedLines.map((group: Line[]) => {
      const heading: Heading = utils.mergeElements<Line, Heading>(
        new Heading(BoundingBox.merge(group.map((l: Line) => l.box))),
        ...group,
      );
      heading.properties.order = group[0].properties.order;
      return heading;
    });
  }

  private noColourFont(font: Font): Font {
    const newFont = new Font(font.name, font.size);
    newFont.weight = font.weight;
    newFont.isItalic = font.isItalic;
    newFont.isUnderline = font.isUnderline;
    newFont.color = null;
    return newFont;
  }

  /**
   * Returns most used font in document
   * @param doc The document to extract heading fonts
   */
  private commonFont(doc: Document): Font {
    return utils.findMostCommonFont(
      doc
        .getElementsOfType<Line>(Line, true)
        .map((l: Line) => l.content)
        .reduce((a, b) => a.concat(b), [])
        .map(w => w.font)
        .filter(f => f !== undefined),
    );
  }

  /**
   * Returns most used line height in document
   * @param doc The document to extract main line height
   */
  /*private commonLineHeight(doc: Document): number {
    const allHeights = doc
      .getElementsOfType<Line>(Line, true)
      .map((l: Line) => l.height)
      .reduce((a, b) => a.concat(b), []);

    const groupedHeights = {};
    allHeights.forEach(i => {
      groupedHeights[i] = (groupedHeights[i] || 0) + 1;
    });

    const mostUsed = Object.keys(groupedHeights).sort(
      (a, b) => groupedHeights[b] - groupedHeights[a],
    );

    return Number(mostUsed[0]);
  }*/

  private pageOtherElements(page: Page): Element[] {
    return page.elements.filter(element => !(element instanceof Paragraph));
  }

  private getElementsWithParagraphs(elements: Element[], withParagraphs: Element[]): Element[] {
    elements.forEach(element => {
      const paragraphs = this.getParagraphsInElement(element);
      if (paragraphs.length > 0) {
        withParagraphs.push(element);
      } else if (Array.isArray(element.content)) {
        element.content.forEach(child => {
          if (child.content as Element[]) {
            this.getElementsWithParagraphs(child.content as Element[], withParagraphs);
          }
        });
      }
    });
    return withParagraphs;
  }

  private getParagraphsInElement(element: Element): Paragraph[] {
    if (Array.isArray(element.content)) {
      return element.content
        .filter(item => item instanceof Paragraph)
        .map(paragraph => paragraph as Paragraph);
    }
    return [];
  }

  private pageParagraphs(page: Page): Paragraph[] {
    return page.getElementsOfType<Paragraph>(Paragraph, false);
  }
  /**
   * Returns an array of fonts to be used for heading detection
   * @param doc The document to extract heading fonts
   */
  private heagingFonts(doc: Document): Font[] {
    const allWords = doc.getElementsOfType<Word>(Word, true);
    // console.log('Total Words ' + allWords.length.toString());
    // const allLines = doc.getElementsOfType<Line>(Line, true).filter(line => line.isUniqueFont());
    // const allFonts = allLines.map(line => this.noColouredFont(line.getMainFont()));
    const allFonts = [...allWords.map(w => this.noColourFont(w.font)).filter(f => f !== undefined)];

    let uniqueFonts: Font[] = [];
    allFonts.forEach(font => {
      if (uniqueFonts.filter(f => f.isEqual(font)).length === 0) {
        uniqueFonts.push(font);
      }
    });

    // console.log('Total Fonts ' + uniqueFonts.length.toString());
    uniqueFonts = uniqueFonts.filter(font => {
      const avg = allWords.filter(w => this.noColourFont(w.font).isEqual(font)).length;
      // Only fonts that are used less than 10% of
      // all words will be used to exctract headings
      // TODO: Add a module param to change this value
      return avg / allWords.length < 0.15;
    });

    // console.log('Heading Fonts');
    // console.log(uniqueFonts);
    // console.log('Heading Fonts ' + uniqueFonts.length.toString());
    // .map(font => this.noColouredFont(font)); // Skip fonts that only have diferent color

    return uniqueFonts;
  }

  private headingsDetected(doc: Document): boolean {
    let detected = false;
    doc.pages.forEach(page => {
      if (page.getElementsOfType<Heading>(Heading, true).length > 0) {
        detected = true;
      }
    });
    return detected;
  }

  private computeHeadingLevels(document: Document) {
    const headings: Heading[] = document.getElementsOfType<Heading>(Heading, true);
    const fontInfo = (heading: Heading) => {
      return {
        size: heading.getMainFont().size,
        weight: heading.getMainFont().weight,
        upperCase: utils.isGeneralUpperCase(heading.content),
      };
    };
    // get all heading fonts sorted by size & upperCase
    // TODO: ¿ sort also by weight ?
    const sortedFonts = headings
      /*.filter(h => {
        console.log(
          h.toString() + ' Footer ' + h.properties.isFooter + ' header ' + h.properties.isHeader,
        );
        return !h.properties.isFooter && !h.properties.isHeader;
      })*/
      .map(h => fontInfo(h))
      .sort((a, b) => {
        if (a.size !== b.size) {
          return b.size - a.size;
        }
        return a.upperCase === b.upperCase ? 0 : a.upperCase ? -1 : 1;
      });
    // remove duplicates
    const uniqueSortedFonts = [...new Set(sortedFonts.map(f => JSON.stringify(f)))].map(s =>
      JSON.parse(s),
    );

    const serializeFont = (heading: Heading) => {
      return (
        fontInfo(heading).size + '|' + fontInfo(heading).weight + '|' + fontInfo(heading).upperCase
      );
    };

    // console.log(uniqueSortedFonts);

    headings.forEach(h => {
      const level = uniqueSortedFonts
        .map(f => f.size + '|' + f.weight + '|' + f.upperCase)
        .indexOf(serializeFont(h));
      h.level = level + 1;
    });
  }
}
